<?php
// Own namespace
namespace Org\Mxchange\CoreFramework\User\Guest;

// Import framework stuff
use Org\Mxchange\CoreFramework\Manager\Guest\ManageableGuest;
use Org\Mxchange\CoreFramework\Registry\Registerable;
use Org\Mxchange\CoreFramework\Request\Requestable;
use Org\Mxchange\CoreFramework\User\BaseUser;
use Org\Mxchange\CoreFramework\User\UsernameMissingException;

/**
 * A generic class for handling guests
 *
 * @author		Roland Haeder <webmaster@shipsimu.org>
 * @version		0.0.0
 * @copyright	Copyright (c) 2007, 2008 Roland Haeder, 2009 - 2021 Core Developer Team
 * @license		GNU GPL 3.0 or any newer version
 * @link		http://www.shipsimu.org
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 */
class Guest extends BaseUser implements ManageableGuest, Registerable {
	// Exceptions
	const EXCEPTION_USERNAME_NOT_FOUND   = 0x170;
	const EXCEPTION_USER_EMAIL_NOT_FOUND = 0x171;
	const EXCEPTION_USER_PASS_MISMATCH   = 0x172;
	const EXCEPTION_USER_NOT_GUEST       = 0x173;

	/**
	 * Protected constructor
	 *
	 * @return	void
	 */
	private function __construct () {
		// Call parent constructor
		parent::__construct(__CLASS__);
	}

	/**
	 * Creates a user by a given request instance
	 *
	 * @param	$requestInstance	An instance of a Requestable class
	 * @return	$userInstance		An instance of this user class
	 * @todo	Add more ways over creating user classes
	 */
	public static final function createGuestByRequest (Requestable $requestInstance) {
		// Determine if by email or username
		if (!is_null($requestInstance->getRequestElement('username'))) {
			// Username supplied
			$userInstance = self::createGuestByUserName($requestInstance->getRequestElement('username'));
		} elseif (!is_null($requestInstance->getRequestElement('email'))) {
			// Email supplied
			$userInstance = self::createGuestByEmail($requestInstance->getRequestElement('email'));
		} else {
			// Unsupported mode
			$userInstance = new Guest();
			$userInstance->debugBackTrace('More ways of creating user classes are needed here.');
			exit();
		}

		// Return the prepared instance
		return $userInstance;
	}

	/**
	 * Creates an instance of this user class by a provided username. This
	 * factory method will check if username is already taken and if not so it
	 * will throw an exception.
	 *
	 * @param	$userName		Username we need a class instance for
	 * @return	$userInstance	An instance of this user class
	 * @throws	UsernameMissingException	If the username does not exist
	 * @throws	UserNoGuestException		If the user is no guest account
	 */
	public static final function createGuestByUsername ($userName) {
		// Get a new instance
		$userInstance = new Guest();

		// Set the username
		$userInstance->setUserName($userName);

		// Check if username exists
		if ($userInstance->ifUsernameExists() === false) {
			// Throw an exception here
			throw new UsernameMissingException(array($userInstance, $userName), self::EXCEPTION_USERNAME_NOT_FOUND);
		} elseif ($userInstance->isGuest() === false) {
			// Sanity check on 'guest' status failed
			throw new UserNoGuestException(array($userInstance, $userName), self::EXCEPTION_USER_NOT_GUEST_STATUS);
		}

		// Return the instance
		return $userInstance;
	}

	/**
	 * Creates an instance of this user class by a provided email address. This
	 * factory method will not check if email address is there.
	 *
	 * @param	$email			Email address of the user
	 * @return	$userInstance	An instance of this user class
	 */
	public static final function createGuestByEmail ($email) {
		// Get a new instance
		$userInstance = new Guest();

		// Set the username
		$userInstance->setEmail($email);

		// Return the instance
		return $userInstance;
	}

	/**
	 * Updates the last activity timestamp and last performed action in the
	 * database result.
	 *
	 * @param	$requestInstance	A requestable class instance
	 * @return	void
	 */
	public function updateLastActivity (Requestable $requestInstance) {
		// No activity will be logged for guest accounts
	}

}
